# Copyright 2007, Sjoerd de Vries
# This file is part of the Spyder main program (a.k.a the Spyder kernel)
# For licensing information, see LICENSE.txt 

import os, zlib, re, sys, imp, compiler
from copy import copy, deepcopy
from Spydercompile import Spydercompile, validvar

commentmatch = re.compile(r'(?:^(?:.*?)##(?:.*?)$)', re.MULTILINE)
brackmatch = re.compile(r'\[[^\[\]]*?\]')
commamatch = re.compile(r'[ \n]*,[ \n]*([^ \n]|$)')
quotematch = re.compile(r'(?<![\\])(([\"\']).*?(?<![\\])\2)')

var = {}
alltypes = []

def changetempdir(tempdir):
  if not os.path.exists(tempdir):
    dirs = [a for a in os.path.abspath(tempdir).split(os.sep) if len(a) > 0]
    currdir = ""
    if dirs[0].endswith(":"):
      currdir = dirs[0]
      dirs = dirs[1:]
    for d in dirs:
      currdir += os.sep + d
      if not os.path.exists(currdir): os.mkdir(currdir)
  os.chdir(tempdir)

def dictparse(s,mask): #adapted from core/parse.py
    if len(s) == 0: return {}
    #mask all sub-lists
    mask0 = mask
    while 1:
        p = brackmatch.finditer(mask0)
        pos = 0
        mask2 = ""
        for pp in p:
            mask2 += mask0[pos:pp.start()] + "!" * (pp.end()-pp.start())
            pos = pp.end()
        mask2 += mask0[pos:]
        if pos == 0: break
        mask0 = mask2
    #split; we can't use split() because of the mask (we need start/end)
    elements = []
    p = commamatch.finditer(mask2)
    pos = 0
    for pp in p:
        elements.append( [s[pos:pp.start()], mask[pos:pp.start()], mask2[pos:pp.start()]] )
        pos = pp.end() - 1
    s = s.rstrip()
    if len(s[pos:].strip()) > 0: elements.append([s[pos:], mask[pos:len(s)], mask2[pos:len(s)]])

    pairs = {}
    for e in elements:
        if e[0].startswith('\n'):
            e[0] = e[0][len('\n'):].lstrip()
            e[1] = e[1][len(e[1])-len(e[0]):]
            e[2] = e[2][len(e[2])-len(e[0]):]
        if len(e[0]) == 0: continue        
        v = e[0]
        m = e[1]
        m2 = e[2]
        v =v.lstrip()
        m = m[len(m)-len(v):]
        m2 = m2[len(m2)-len(v):]      
        bm = m.find('[')
        bm2 = m.rfind(']')
        ee = m2.find(":")
        if ee > -1: #pair
            k = v[:ee].rstrip()
            v = v[ee+1:]
            m = m[ee+1:]
            v =v.lstrip()
            m = m[len(m)-len(v):]
            if (k[0] != "\"" or k[-1] != "\"") and (k[0] != "'" or k[-1] != "'"):
                raise SyntaxError("Malformed dictionary key %s" % k)
            k = k[1:-1]
            if k in pairs:
                raise SyntaxError("Double key %s" % k)
            bm = m.find('[')
            bm2 = m.rfind(']')
            
            if v[0] in ('[','('): #list
                v = v[1:-1]
                m = m[1:-1]
                pairs[k] = listparse(v,m)
            elif v[0] == '{': #dict
                v = v[1:-1]
                m = m[1:-1]
                pairs[k] = dictparse(v,m)
            elif bm > -1: #function + args
              print s
              print m
              raise Exception
            else: #value
              assert (v[0] == v[-1] == "'") or (v[0] == v[-1] == '"')
              pairs[k] = eval("str(%s)" % v)
        else:
            raise SyntaxError("Malformed dictionary pair %s" % e[0])
    return pairs

def listparse(s,mask): #adapted from core/parse.py
    if len(s) == 0: return []
    #mask all sub-lists
    mask0 = mask
    while 1:
        p = brackmatch.finditer(mask0)
        pos = 0
        mask2 = ""
        for pp in p:
            mask2 += mask0[pos:pp.start()] + "!" * (pp.end()-pp.start())
            pos = pp.end()
        mask2 += mask0[pos:]
        if pos == 0: break
        mask0 = mask2
    #split; we can't use split() because of the mask (we need start/end)
    elements = []
    p = commamatch.finditer(mask2)
    pos = 0
    for pp in p:
        elements.append( [s[pos:pp.start()], mask[pos:pp.start()], mask2[pos:pp.start()]] )
        pos = pp.end() - 1
    s = s.rstrip()
    if len(s[pos:].strip()) > 0: elements.append([s[pos:], mask[pos:len(s)], mask2[pos:len(s)]])

    values = []
    for e in elements:
      if e[0].startswith('\n'):
          e[0] = e[0][len('\n'):].lstrip()
          e[1] = e[1][len(e[1])-len(e[0]):]
      if len(e[0]) == 0: continue
      v = e[0]
      m = e[1]
      v =v.lstrip()
      m = m[len(m)-len(v):]  
      bm = m.find('[')
      bm2 = m.rfind(']')      
      if v[0] in ('[','('): #list
          v = v[1:-1]
          m = m[1:-1]
          values.append(listparse(v,m))
      elif v[0] == '{': #dict
          v = v[1:-1]
          m = m[1:-1]
          values.append(dictparse(v,m))
      elif bm > -1: #function + args
        raise Exception
      else: #value
        assert v[0] == v[-1] 
        assert v[0] in ( "'", '"') 
        values.append(eval("str(%s)" % v))
    return values



class SystemType:
  def __init__(self):
    self.definition = None
    self.support = set()

class general(object):
  pass

reservedswitches = (
  "NO_COMPILE",
  "TEMPDIR",
)

Spyder = general()

inheritedtypes = {}
def reset():
  global unimethods, unimethodlist, macros, statements, inheritedtypes, Spyder, corechecksum
  global converters, concatenators
  unimethods = {}
  var = {}
  unimethodlist = []
  macros = set()
  statements = {}
  converters = []
  concatenators = []
  inheritedtypes = {}
  corechecksum = 0  
  

def defineunimethod(name, pointer):
  if name in unimethods:
    raise Exception("system error: duplicate definition of unimethod %s" % name)
  unimethods[name] = pointer
  unimethodlist.append(name)
def definemacro(pointer):
  macros.add(pointer)
def definestatement(name, pointer):
  if name in statements:
    raise Exception("system error: duplicate definition of statement %s" % name)
  statements[name] = pointer  

def parse(lines, switches, switchlist,spyderdir):
  switchvars = []
  currnamespace = ""
  currswitch = ""
  spacing = [0,]
  tabin = False
  linenr = 0
  toplevelfiles = []
  tmpvars = [{}]
  metavars = {"nocripple":"off"}
  if spyderdir != None: tmpvars[-1]["spyderdir"] = spyderdir  
  
  while len(lines) > 0:
    l, lines = lines[0], lines[1:]
    l = l.replace('\r\n', "")
    l = l.replace('\n', "")
    linenr += 1
    if l.strip().startswith("@"):
      ll = l.strip()[1:].split("=")
      if len(ll) == 2:
        metavars[ll[0].strip()] = ll[1].strip()
      else:
        raise Exception("Invalid metavariable statement: " + l)
    comment = l.find("#")
    if comment > -1: 
      special = False
      if metavars["nocripple"] == "on":
        ll = l.strip()
        indent = len(l) - len(ll) 
        if ll.startswith("##"):
          ll = ll[2:].strip()
          if ll.startswith("NOCRIPPLED"):
            header = " " * indent
            if indent == 0: 
              header = "##"
            else: special = True
            l = header + ll[len("NOCRIPPLED"):].strip()
      if special == False:
        l = l[:comment]
    if len(l.lstrip()) == 0: continue
    dollarstart = 0
    while 1:     
      dollar = l.find("$",dollarstart)
      if dollar == -1: break
      for endvar in range(dollar+1,len(l)):
        if l[endvar].isalnum() == False or l[endvar] == " ":
          break
      else: endvar = len(l)
      varnam = l[dollar+1:endvar]
      if varnam not in tmpvars[-1]:
        raise Exception("Unknown variable: '%s'" % varnam)
      l = l[:dollar] + tmpvars[-1][varnam] + l[endvar:]
      dollarstart = dollar + len(tmpvars[-1][varnam])
    currspacing = len(l) - len(l.lstrip())
    if tabin == True:
      if currspacing < spacing[-1]:
        raise Exception("Indentation error in line %d" % linenr)
      elif currspacing == spacing[-1]:
        spacing.append(999999)
        tabout = True
      else:
        spacing.append(currspacing)
        tabin = False
        tabout = False
    else: tabout = True
    
    if tabout == True:
      if currspacing > spacing[-1]:
        raise Exception("Indentation  error in line %d" % linenr)
      if currspacing < spacing[-1]:
        counter = -1
        for n in range(len(spacing)-2,-1,-1):
          if currspacing == spacing[n]:
            counter = len(spacing) - 2 - n
            break
        if counter == -1: raise Exception("Indentation error in line %d" % linenr)
        for n in range(counter+1):
          if currnamespace != "": 
            switches[currswitch].append("**")
            currnamespace = ""
          elif currswitch != "":
            currswitch = ""
            if len(tmpvars) > 2:
              for k in tmpvars[-1]:
                if k not in tmpvars[-2]: tmpvars[-2][k] = tmpvars[-1][k]
            tmpvars.pop()
          spacing.pop()
    l = l.strip()
    if l[0] != "*" and l[-1] == ":" :
      if currnamespace != "": raise Exception("System error: %s:\n Modules may not be nested" % l)
      currnamespace = l[:-1].rstrip()
      tmpvars.append(dict(tmpvars[-1]))
      tabin = True
      switches[currswitch].append(l)
      continue
    if l[0] == "*":
      ll = l.split()
      found = False
      for com in ("*path", "*delpath", "*delfile", "*replacefile", "*include", "*import", "*export"):
        if ll[0] == com:
          if currnamespace == "": raise Exception("System error: %s:\n %s only allowed in modules" % (l, ll[0]))
          found = True
      for com in ("*enable", "*disable", "*select", "*deselect"):
        if ll[0] == com:
          found = True
          
      if found == False:
        if ll[0] == "*patch" :
          if currnamespace != "": raise Exception("System error: %s:\n %s only allowed outside modules" % (l, ll[0]))
          currnamespace = l[len(ll[0]):].rstrip()
          if currnamespace[-1] == ":": currnamespace = currnamespace[:-1]
          tmpvars.append(dict(tmpvars[-1]))
          tabin = True
        elif ll[0] == "*switch":
          if currswitch != "": raise Exception("System error: %s:\switches cannot be nested" % l)
          if currnamespace != "": raise Exception("System error: %s:\n %s only allowed outside modules" % (l, ll[0]))
          currswitch = l[len(ll[0]):].lstrip().upper()
          if currswitch[-1] == ":": currswitch = currswitch[:-1]
          if currswitch in reservedswitches: raise Exception("System error: switchname %s has been reserved" % currswitch)
          if currswitch not in switches:
            switches[currswitch] = []
            switchlist.append(currswitch)
          tabin = True
          continue
        else:
          raise Exception("unknown command %s" % l)        
      switches[currswitch].append(l)
      continue
    if l.find("=") > -1:
      varstat = l.split('=')
      varnam=varstat[0].strip()
      varval=varstat[1].strip()
      if len(varnam.split()) > 1:
        raise Exception("system error: invalid command\n%s" % l)
      tmpvars[-1][varnam] = varval
      switchvars.append((currswitch,varnam,varval))
      continue
    if len(l.split()) > 1:
      raise Exception("system error: invalid command\n%s" % l)
    if currnamespace == "": 
      if l.upper().endswith(".FLY"):
        try:
          dat = open(l).readlines()          
        except:
          raise Exception("Cannot read %s" % l)
        flydir = os.path.dirname(os.path.abspath(l))        
        lines = dat + lines
        lines = ["installdir = %s\n" % flydir] + lines
      else:      
        toplevelfiles.append(l)
      continue
    switches[currswitch].append(l)
    continue
  for n in range(len(spacing)):
    if currnamespace != "": 
      switches[currswitch].append("**")
      currnamespace = ""
    elif currswitch != "":
      currswitch = ""
  return toplevelfiles, switchvars
    
def assess_enabling(a, spyderdir, has_command_tempdir = None):
  global var, has_tempdir, systempdir
  if spyderdir != None: var["spyderdir"] = spyderdir  
  if has_command_tempdir == None:
    has_command_tempdir = has_tempdir
  first = (len(a.checked) == 0)
  for swvar in a.switchvars:    
    sw,varnam,varval = swvar
    if sw == a.currswitch:
      var[varnam] = varval
      if varnam.lower() == "tempdir":
        if not has_command_tempdir: 
          systempdir = varval
          changetempdir(systempdir)
        has_tempdir = True
  
  for l in a.switches[a.currswitch]:
    if l[0] == "*":
      ll = l.split()
      if ll[0] == "*enable":
        ll = l.split()
        switch = l[len(ll[0]):].lstrip().upper()
        if switch in a.switches: #No checking if switch is defined!
          a.enabled.add(switch)
      if ll[0] == "*disable":
        ll = l.split()
        switch = l[len(ll[0]):].lstrip().upper()
        if switch in a.enabled: #No checking if switch is defined!
          a.enabled.remove(switch)
  a.checked.add(a.currswitch)
  if first:
    change = True
    while change == True:
      change = False
      for switch in a.switchlist:
        if switch not in a.checked and switch in a.enabled:
          change = True
          a.currswitch = switch              
          assess_enabling(a,spyderdir,has_command_tempdir)

def assemble(namespaces, lines):
  currnamespace = ""
  ignore = False
  disabled = set()  
  while len(lines) > 0:
    l = lines[0]
    lines = lines[1:]
    l = l.strip()
    if l[0] != "*" and l[-1] == ":" :
      if currnamespace != "": raise Exception("System error: %s:\n Modules may not be nested" % l)
      currnamespace = l[:-1].rstrip()
      if currnamespace in namespaces:
        raise Exception("System error: Module %s already defined" % (currnamespace))
      if currnamespace not in disabled:
        namespaces[currnamespace] = general()
        nam = namespaces[currnamespace]
        nam.order = len(namespaces)
        nam.name = currnamespace
        nam.iscore = None
        nam.paths = []
        nam.files = []
        nam.codefiles = []
        nam.dependent = set()
        nam.defined = {}
        nam.supported = set()
        nam.included = set()
        nam.precompiled = None
        nam.compiled = False
        nam.checksum = None
        nam.checksum_verified = False
        nam.source = ""
        nam.filecompiled = 0
        nam.id = len(namespaces)
        nam.imports = []
        nam.exports = []
      continue
    if ignore == True and ll[0] != "**": continue
    if l[0] == "*":
      ll = l.split()
      if ll[0] == "**":
        currnamespace = ""
        ignore = False
      elif ll[0] == "*path":
        if currnamespace == "": raise Exception("System error: %s:\n %s only allowed in modules" % (l, ll[0]))
        p = l[len(ll[0]):].lstrip()
        nam.paths.append(p)
      elif ll[0] == "*import":
        if currnamespace == "": raise Exception("System error: %s:\n %s only allowed in modules" % (l, ll[0]))
        p = l[len(ll[0]):].lstrip()
        nam.imports.append(p)
      elif ll[0] == "*export":
        if currnamespace == "": raise Exception("System error: %s:\n %s only allowed in modules" % (l, ll[0]))
        p = l[len(ll[0]):].lstrip()
        nam.exports.append(p)
      elif ll[0] == "*delpath":
        if currnamespace == "": raise Exception("System error: %s:\n %s only allowed in modules" % (l, ll[0]))
        p = l[len(ll[0]):].lstrip()
        if p in nam.paths:
          nam.paths.remove(p)
      elif ll[0] == "*delfile":
        if currnamespace == "": raise Exception("System error: %s:\n %s only allowed in modules" % (l, ll[0]))
        p = l[len(ll[0]):].lstrip()
        if p in nam.files:
          nam.files.remove(p)
      elif ll[0] == "*replacefile":
        if currnamespace == "": raise Exception("System error: %s:\n %s only allowed in modules" % (l, ll[0]))
        p = l[len(ll[0]):].lstrip()
        pp = p.split()
        if len(pp) != 2:
          raise Exception("System error: Malformed *replacefile: %s" % (l))
        pp[0] = pp[0].strip()
        pp[1] = pp[1].strip()        
        if pp[0] in nam.files:
          nam.files[nam.files.index(pp[0])] = pp[1]
      elif ll[0] == "*include":
        if currnamespace == "": raise Exception("System error: %s:\n %s only allowed in modules" % (l, ll[0]))        
        nam.included.add(ll[1])          
      elif ll[0] == "*patch" :
        if currnamespace != "": raise Exception("System error: %s:\n %s only allowed outside modules" % (l, ll[0]))
        currnamespace = l[len(ll[0]):].lstrip()
        if currnamespace[-1] == ":": currnamespace = currnamespace[:-1]
        if currnamespace not in namespaces:
          if currnamespace in disabled:
            ignore = True
          else:
            raise Exception("System error: Module %s cannot be patched since it is undefined" % (currnamespace))
        nam = namespaces[currnamespace]
      elif ll[0] == "*select": #modules that are defined later are still added!
        delspaces = []
        for nam in namespaces:
          if nam not in ll[1:]:
            disabled.add(nam)
            delspaces.append(nam)
        for nam in delspaces: namespaces.pop(nam)
      elif ll[0] == "*deselect":
        delspaces = []        
        for nam in namespaces:
          if nam in ll[1:]:
            disabled.add(nam)
            delspaces.append(nam)
        for nam in delspaces: namespaces.pop(nam)        
      elif ll[0] == "*enable" :
        continue
      elif ll[0] == "*disable" :
        continue      
      else:
        raise Exception("unknown command %s" % l)
      continue
    if currnamespace == "": raise Exception
    psplit = l.split(".")
    if len(psplit) != 2: raise Exception("System error: file name %s should contain exactly one dot" % l)
    ext = psplit[1].upper()
    if ext == "SPY" or ext == "PY":
      nam.files.append(l)
    elif ext == "FLY":
      if not os.path.exists(l):
        raise Exception("System error: file name %s does not exist" % l)
      lines = open(l).readlines() + lines
    else:
      raise Exception("System error: file name %s must be Spyder or Python file" % l)
    
    

def analyze_python_file(system, fil, lines, namespace, iscompiled, make_definitions=True):
  for l in lines:
    l = l.strip()
    if len(l) < 2: continue
    if l[0] != "#": break
    if l[:2] != "##": continue
    l = l[2:].strip()
    l1 = l.split()[0].rstrip()
    l2 = l[len(l1):].lstrip()
    l2 = l2.split(",")
    for n in range(0,len(l2)): l2[n] = l2[n].strip()
    if l1 == "DEFINE":
      if namespace.iscore == True:
        raise Exception("System error in %s: module may not contain core files and DEFINE metastatements" % (namespace.name))
      namespace.iscore = False            
      for ll in l2:
        if not validvar(ll):
          raise Exception("System error in file %s: defined type %s must start with a capital" % (fil,ll))
        if ll == "Spyder" or ll == "Delete": raise Exception("System error in file %s: defined type may not be 'Spyder'" % (fil))
        if ll in system and system[ll].definition != None and system[ll].definition != namespace.name:
          raise Exception("System error in file %s: type %s already defined in %s" % (fil, ll, system[ll].definition))
        if ll not in system: system[ll] = SystemType()
        if make_definitions == True: system[ll].definition = namespace.name
        namespace.defined[ll] = ""
    elif l1 == "SUPPORT":
      if namespace.iscore == True:
        raise Exception("System error in %s: module may not contain core files and SUPPORT metastatements" % (namespace.name))
      namespace.iscore = False      
      for ll in l2:
        if ll == "None": continue
        if not validvar(ll):
          raise Exception("System error in file %s: supported type %s must start with a capital" % (fil,ll))
        if ll == "Spyder" or ll == "Delete": raise Exception("System error in file %s: defined type may not be 'Spyder' or 'Delete'" % (fil))
        if ll not in system: system[ll] = SystemType()
        system[ll].support.add(namespace.name)
        namespace.supported.add(ll)
    elif l1 == "CHECKSUM":
      if iscompiled == False: 
          raise Exception("System error in file %s: only compiled file may contain CHECKSUM metastatement" % (fil))            
      if len(l2) < 1:
        raise Exception("System error in file %s: malformed CHECKSUM metastatement" % (fil))
      checksum = l[len(l1):].lstrip()
      namespace.checksum = checksum #checksum of the SOURCE plus CORE SOURCE, not of the file itself!
    elif l1 == "NEEDED":
      if namespace.iscore == True:
        raise Exception("System error in %s: module may not contain core files and NEEDED metastatements" % (namespace.name))
      namespace.iscore = False            
      for ll in l2:
        if not validvar(ll):
          raise Exception("System error in file %s: needed type %s must start with a capital" % (fil,ll))
        if ll == "Spyder" or ll == "Delete": raise Exception("System error in file %s: needed type may not be 'Spyder'" % (fil))
        namespace.neededtypes.add(ll)
    elif l1 == "DEPENDENT":
      if iscompiled == False: 
          raise Exception("System error in file %s: only compiled file may contain NEEDED metastatement" % (fil))                  
      if namespace.iscore == True:
        raise Exception("System error in %s: module may not contain core files and DEPENDENT metastatements" % (namespace.name))
      namespace.iscore = False
      l2 = l[len(l1):].split()
      if len(l2) != 2:
        raise Exception("System error in file %s: malformed DEPENDENT metastatement" % (fil))
      checksum = l2[1].strip()
      namespace.dependent.add((l2[0], checksum))
    elif l1 == "CORE":
        if namespace.iscore == False:
          raise Exception("System error in %s: module may not contain core files and Spyder files/meta statements" % (namespace.name))
        namespace.iscore = True
    else: raise Exception("System error in file %s:\n Unrecognized metacommand %s" % (fil, l1))
  if iscompiled == True and namespace.iscore != True: namespace.iscore = False

anyfound = False
def compile_core_namespace(system, namespace):
  global corechecksum, anyfound
  source = ""
  path = ""
  if len(namespace.paths) > 0: path = namespace.paths[0]
  if has_tempdir: fil = namespace.name
  else: fil = os.path.join(path, namespace.name)
  fil += "_MODULE_spy.py"
  namespace.compiledfile = fil
  namespace.precompiled = False
  namespace.neededtypes = set()
  if os.path.exists(fil):
    f = open(fil)
    d = f.read().replace('\r\n','\n').rstrip('\n')
    lines = d.split("\n")
    namespace.source = d
    f.close()
    analyze_python_file(system, fil, lines, namespace, True, False)
    namespace.precompiled = None

  for filename in namespace.files:
    ok = False
    for p in namespace.paths:
      fil = os.path.join(p, filename)
      if os.path.exists(fil):
        f = open(fil)
        d = f.read().replace('\r\n','\n').rstrip('\n')
        lines = d.split('\n')
        source += d + '\n'
        namespace.codefiles.append(fil)
        f.close()
        ok = True
        if filename.upper().endswith("SPY"):
          if namespace.iscore == True:
            raise Exception("System error in %s: module may not contain core files and Spyder files" % (namespace.name))
          namespace.iscore = False
        elif filename.upper().endswith("PY"):
          if namespace.precompiled == False:
            analyze_python_file(system, fil,lines, namespace, False)
        else:
          raise Exception("Only .spy and .py files can be present in modules")
        break
    if ok == False: 
      if anyfound == False:
        raise Exception("Spyder could not locate its core functions. Your Spyder install seems to be broken")
      else: raise Exception("System error: cannot find %s in module %s" % (filename,namespace.name))
    anyfound = True
  if namespace.iscore == True:
    checksum = hex(zlib.adler32(source))        
    if namespace.precompiled == None:
      namespace.precompiled = True
      if checksum != namespace.checksum:
        namespace.precompiled = False
    corechecksum = zlib.adler32(source, corechecksum)
    if namespace.precompiled == False:
      source = ""
      #print >> sys.stderr, "compiling core module '%s'..." % namespace.name 
      f = open(namespace.compiledfile, "w")
      f.write("## CORE\n")
      f.write("## CHECKSUM %s\n\n" % checksum)
      for fil in namespace.codefiles:
        l = "execfile(r\"%s\")\n" % os.path.abspath(fil)
        source += l
        f.write(l)
      
      f.close()

      filelist = namespace.compiledfile[:-3] + "_file.list"
      f = open(filelist, "w")
      for l in namespace.codefiles:
        f.write(l + '\n')
      f.close()
    else:
      namespace.codefiles = []
      f = open(namespace.compiledfile[:-3] + "_file.list")
      d = f.readlines()
      for l in d:
        ll = l.rstrip('\n')
        namespace.codefiles.append(ll)
      f.close()
      

    #print >> sys.stderr, "executing core module '%s'..." % namespace.name
    shell = {'Spyder':Spyder}
    Spyder.__currpaths__ = namespace.paths    
    head, tail = headtailcode(namespace.name, namespace.paths, namespace.imports,namespace.exports) 
    #source = head + '\n' + source + tail + '\n'
    try:
      exec(head  % dict(shell="shell"))
      exec(source, shell)
      exec(tail % dict(shell="shell"))
    except:
      #dump = open("dump.py", "w")
      #dump.write(source)
      #dump.close()
      raise
    for var in shell:
      if validvar(var) and var != 'Spyder':
        inheritedtypes[var] = {}
    namespace.source = source
    namespace.compiled = True
    namespace.calledtypes = set()
  else: namespace.defined = {}
  
def compile_namespace(system, namespaces, namespace):  
  if namespace.checksum_verified == False:
    compiledfiles = set()
    fil = namespace.compiledfile    
    if not (os.path.exists(fil[:-3] + "_called.types") and os.path.exists(fil[:-3] + "_defined.types") and os.path.exists(fil[:-3] + "_file.list")):
      namespace.precompiled = False
      namespace.supported = set()
      namespace.defined = {}
      namespace.source = ""
      namespace.dependent = set()
    else:
      f = open(fil[:-3] + "_file.list")
      d = f.readlines()
      f.close()
      for l in d:
        if l.strip().endswith("_spy.py") and not os.path.exists(l.strip()):
          namespace.precompiled = False
          namespace.supported = set()
          namespace.defined = {}
          namespace.source = ""
          namespace.dependent = set()
          break          
      else:
        for l in d:
          ll = l.rstrip('\n')
          compiledfiles.add(ll)    
    foundfiles = set()
    source = ""
    for n in range(0,len(namespace.files)):
      filename = namespace.files[n]
      ok = False
      for p in namespace.paths:
        fil = os.path.join(p, filename)
        if os.path.exists(fil):
          ok = True
          if filename.upper().endswith("SPY"):
            if has_tempdir:   
              mdir = namespace.name
              if not os.path.exists(mdir):
                os.mkdir(mdir)
              foundfiles.add(mdir + os.sep + os.path.basename(fil[:-4])+"_spy.py")
            else: 
              foundfiles.add(fil[:-4]+"_spy.py")
            f = open(fil)
            d = f.read().replace('\r\n','\n').rstrip('\n')
            source += d + '\n\n'
            f.close()
          else:
            foundfiles.add(fil)
            f = open(fil)
            d = f.read().replace('\r\n','\n').rstrip('\n')
            source += d + '\n\n'
            lines = d.split('\n')
            f.close()
          namespace.files[n] = fil
          break
      if ok == False: raise Exception("System error: cannot find %s" % filename)        
    checksum = hex(zlib.adler32(source, corechecksum))
    if namespace.precompiled == None:
      if checksum != namespace.checksum or len(foundfiles ^ compiledfiles) > 0:
        namespace.precompiled = False
        namespace.supported = set()
        namespace.defined = {}
        namespace.source = ""
        namespace.dependent = set()
      else: namespace.precompiled = True 
    namespace.checksum = checksum  
    namespace.checksum_verified = True

  if namespace.precompiled == True:
    if not hasattr(namespace,"dependent_copy"):
      namespace.dependent_copy = namespace.dependent
    #check if the dependency namespaces are in the system, and identical in checksum
    dep = list(namespace.dependent_copy)
    namespace.dependent_copy = set()
    for n in range(0, len(dep)):
      d = dep[n]
      ok = True
      if d[0] not in namespaces:
        ok = False
      elif namespaces[d[0]].compiled == False:
        namespace.dependent_copy.add(d)
      elif namespaces[d[0]].checksum != d[1]:
        ok = False
      if ok == False:
        namespace.precompiled = False
        namespace.supported = set()
        namespace.defined = {}
        namespace.neededtypes = set()              
        namespace.dependent = set()
        namespace.source = ""
        break
    if ok == True: namespace.dependent = namespace.dependent_copy
    if namespace.precompiled == True and len(namespace.dependent) == 0: #every test has been passed
      analyze_python_file(system, namespace.compiledfile, namespace.source.splitlines(), namespace, True)
      namespace.compiled = True
      
      f = open(namespace.compiledfile[:-3] + "_defined.types")
      d = f.read().replace('\r\n','\n')
      f.close()
      for l in d.splitlines():
        sep = l.index(" ")
        typ, s = l[:sep], l[sep+1:]
        namespace.defined[typ] = {}
        mask = ""
        pos = 0
        p = quotematch.finditer(s)
        for pp in p:
            mask += s[pos:pp.start()] + "*" * (pp.end()-pp.start())
            pos = pp.end()
        mask += s[pos:]
        mask = mask.replace('{', '[').replace('}',']')
        mask = mask.replace('(', '[').replace(')',']')
        mask = mask.replace(r'\\', '**')
        m = mask
        bm = m.find('[')
        if bm > -1:
          bm2 = m.rfind(']')
          cm = m.find(',')
          s = s[bm+1:bm2]
          m = m[bm+1:bm2]
          namespace.defined[typ] = dictparse(s,m)
      f = open(namespace.compiledfile[:-3] + "_called.types")
      d = f.read().replace('\r\n','\n')
      f.close()
      namespace.calledtypes = set(d.splitlines())

      namespace.codefiles = []
      f = open(namespace.compiledfile[:-3] + "_file.list")
      d = f.readlines()
      for l in d:
        ll = l.rstrip('\n')
        namespace.codefiles.append(ll)
      f.close()
      for var in namespace.defined:
        if var not in system: 
          system[var] = SystemType()
        if system[var].definition != namespace.name and system[var].definition != None:
          raise Exception("compile error: type %s twice defined in modules %s and %s" % (var, namespace.name, system[var].definition))
        system[var].definition = namespace.name
  if namespace.precompiled == False:
    #sys.exit()
    change = True
    #filedefined = {}
    while change:
      change = False
      newnt = copy(namespace.neededtypes)
      for nt in namespace.neededtypes:
        if nt not in system:
          system[nt] = SystemType()
        if system[nt].definition != None:
          depnam = namespaces[system[nt].definition]
          if depnam.id > namespace.id:
            raise Exception("ERROR: Module %s defines type %s, which is dependent on module %s, which is defined later in system.fly\nChange the order of modules" %(namespace.name, nt, depnam.name))
          if depnam.compiled == True:
            namespace.dependent.add((depnam.name, depnam.checksum))
            newnt.remove(nt)
          elif depnam.name == namespace.name:
            newnt.remove(nt)
      namespace.neededtypes = newnt
      if len(namespace.files) == 0:
        namespace.calledtypes = set()
        namespace.compiled = True
      elif len(namespace.neededtypes) == 0:
        filename = namespace.files[namespace.filecompiled]
        f = open(filename)
        s = f.read().replace('\r\n','\n')
        f.close()
        if filename.upper().endswith(".SPY"):
          #print "NAMESPACE", namespace.name
          #print "FILE", filename
          inheritedtypes0 = inheritedtypes.copy()
          for v in namespace.defined:
            inheritedtypes0[v] = namespace.defined[v]
          result, neededtypes, definedtypes, supportedtypes = Spydercompile(Spyder, s, unimethodlist, unimethods, macros, statements, inheritedtypes0, namespace.name, filename)
          #print "RESULT", result
          #print "NEEDEDTYPES", neededtypes
          #print "DEFINEDTYPES", definedtypes
          #print "SUPPORTEDTYPES", supportedtypes
          if result == None:
            namespace.neededtypes = neededtypes
          else:
            result = result.lstrip('\n') + '\n'
            linenr = len(result.splitlines())
            if has_tempdir:
              mdir = namespace.name
              if not os.path.exists(mdir):
                os.mkdir(mdir)
              spynam = mdir + os.sep + os.path.basename(filename)[:-4]+"_spy.py"
            else: spynam = filename[:-4]+"_spy.py"
            spyfile = open(spynam, "w")
            spyfile.write(result)
            spyfile.close()
            namespace.codefiles[namespace.filecompiled] = spynam
            namespace.source += result
            namespace.supported = namespace.supported.union(supportedtypes)
            for var in definedtypes:
              if var in namespace.defined:
                raise Exception("compile error: type %s twice defined in module %s" % (var, namespace.name))
              if var not in system: 
                system[var] = SystemType()
              if system[var].definition != namespace.name and system[var].definition != None:
                raise Exception("compile error: type %s twice defined in modules %s and %s" % (var, namespace.name, system[var].definition))
              system[var].definition = namespace.name
              namespace.defined[var] = definedtypes[var]
            change = True
            namespace.filecompiled += 1
        else:
          #if filename not in filedefined:
          s = s.lstrip('\n') + '\n\n'
          namespace.source += s
          lines = s.split("\n")
          tmpnamespace = general()
          tmpnamespace.name = namespace.name
          tmpnamespace.defined = {}
          tmpnamespace.neededtypes = set()
          tmpnamespace.supported = set()
          tmpnamespace.iscore = False
          analyze_python_file(system, filename,lines, tmpnamespace, False)
          #filedefined[filename] = tmpnamespace.defined
          namespace.supported = namespace.supported.union(tmpnamespace.supported)
          if len(tmpnamespace.neededtypes) > 0:
            namespace.neededtypes = tmpnamespace.neededtypes
            #change = True
            #continue            
          #if filedefined[filename] > 0:
          #if filename.upper().endswith(".SPY"):
          #shell = {'Spyder':Spyder}
          #dump = open("dump.py", "w")
          #dump.write(namespace.source)
          #dump.close()            
          #exec(namespace.source, shell)
          #ok = True
          for var in tmpnamespace.defined:
            if var in namespace.defined:
              raise Exception("compile error: type %s twice defined in module %s" % (var, namespace.name))
            namespace.defined[var] = {}
            if var not in system: 
              system[var] = SystemType()
            if system[var].definition != namespace.name and system[var].definition != None:
              raise Exception("compile error: type %s twice defined in modules %s and %s" % (var, namespace.name, system[var].definition))
            system[var].definition = namespace.name
            #namespace.defined[var] = definedtypes[var]
            namespace.defined[var] = tmpnamespace.defined[var]

              
          change = True                        
          namespace.filecompiled += 1
        if namespace.filecompiled == len(namespace.files) and len(namespace.neededtypes) == 0:
          if "SILENT" not in assesspar.enabled: print >> sys.stderr, "compiling module '%s'" % namespace.name
          #filteredsource0 = commentmatch.split(namespace.source)
          #filteredsource = ""
          #for f in filteredsource0: filteredsource += f
          filteredsource = namespace.source
          f = open(namespace.compiledfile, "w")
          f.write("\n## CHECKSUM %s\n\n" % namespace.checksum)
          for v in namespace.defined:
            f.write("## DEFINE %s\n" % v)
          for v in namespace.supported:
            if v not in namespace.defined:
              f.write("## SUPPORT %s\n" % v)
          for v in namespace.dependent:          
            f.write("## DEPENDENT %s %s\n" % (v[0], v[1]))
          f.write("\n" + filteredsource)
          f.close()
          definedtypes = namespace.compiledfile[:-3] + "_defined.types"
          f = open(definedtypes, "w")
          #f.write(repr(namespace.defined))
          for t in namespace.defined:
            print >>f, t, repr(namespace.defined[t])
          f.close()

          filelist = namespace.compiledfile[:-3] + "_file.list"
          f = open(filelist, "w")
          for l in namespace.codefiles:
            print >> f, l
          f.close()
          namespace.calledtypes = get_called_types(namespace.source)
          calledtypes = namespace.compiledfile[:-3] + "_called.types"
          f = open(calledtypes, "w")
          for t in namespace.calledtypes:
            print >> f, t
          f.close()
          
          namespace.compiled = True
          change = False

  if namespace.compiled == True:
    for v in namespace.defined:
      inheritedtypes[v] = namespace.defined[v]
      Spyder.core.__defined[v] = namespace.defined[v]
     
def visit_node(node, ret):
  nodetype = node.__class__.__name__
  if nodetype == "Name":
    name = node.getChildren()[0]
    name0 = name
    while name0.endswith("Array"): name0 = name0[:-len("Array")]
    if validvar(name0):      
      ret.add(name)
  for child in node.getChildNodes():
    visit_node(child, ret)

def get_called_types(code, raise_syntaxerror = True):
  try:
    try:
      ast = compiler.parse(code)
    except SyntaxError:
      #at any point, it will crash, just depends if we want to let it crash now or later...      
      try:    
        exec(code) 
      except SyntaxError:
        if raise_syntaxerror == False: return set() 
        raise        
      except:
        raise
  except:
    #print "Compilation error, check dump.py to fix it"
    #dump = open("dump.py", "w")
    #dump.write(code)
    #dump.close()
    raise
  ret = set()
  visit_node(ast, ret)
  return ret

def determine_called_namespaces(system, namespaces, name):
  namespace = namespaces[name]
  if hasattr(namespace, "called"): return
  namespace.called = []
  for t in namespace.calledtypes:
    while t.endswith("Array"): t = t[:-len("Array")]      
    if t not in system or system[t].definition not in namespaces:
      raise Exception("System error: module %s needs Spyder type %s and no system module can provide it" % (name,t))
    tt = system[t]
    if namespaces[tt.definition].id > namespace.id:
      raise Exception("System error: module %s needs Spyder type %s and the defining system module %s is later in system.fly\nPlease change the order of modules" % (name,t,tt.definition))
    newnames = set((tt.definition,))
    if hasattr(tt, "supported") and tt.supported != None: newnames = newnames.union(set(tt.supported))
    for n in newnames:
      if n != name:
        determine_called_namespaces(system, namespaces, n)
      if n not in namespace.called:
        for nn in namespaces[n].called:
          if nn not in namespace.called: namespace.called.append(nn)
  namespace.called.append(name)
  for i in namespace.included:
    if i not in namespace.called: namespace.called.append(i)

def headtailcode(modulename, paths, imports, exports):
  ret1 = "Spyder.core.__currpaths = %s\n" % repr(paths)
  ret1 += "if \"%s\" not in Spyder.__dict__: Spyder.%s = Spyder.core.namespace()\n" % (modulename, modulename)  
  for i in imports:
    txt = "\""+ i + "\",\"" + i + "\""
    ii = i.split()
    
    nam = ii[0]
    val = ii[0]
    if (len(ii) != 1 and (len(ii) != 3 or ii[1] != "as")) or validvar(ii[0]): 
      raise Exception("Malformed import statement in %s: *import %s" % (modulename,i))
    if len(ii) == 3:
      txt = "\"" + ii[0] + "\",\"" + ii[2] + "\""
      val = ii[2]
    if val.find(".") > -1:
      val = val[:val.index(".")]
    ret1 += "Spyder.core_init.__fullimport(%s, %%(shell)s)\n" % (txt)
    ret1 += "Spyder.%s.%s = %%(shell)s[\"%s\"]\n" % (modulename, val, val)
  ret2 = ""
  for i in exports:
    ii = i.split()
    nam = ii[0]
    val = ii[0]
    if (len(ii) != 1 and (len(ii) != 3 or ii[1] != "as")) or validvar(ii[0]): 
      raise Exception("Malformed export statement in %s: *export %s" % (modulename,i))
    if len(ii) == 3:
      val = ii[2]
    if nam.find(".") > -1:
      raise Exception("Malformed export statement in %s: *export %s" % (modulename,i))
    ret2 += "Spyder.%s.%s = %%(shell)s[\"%s\"]\n" % (modulename, val, nam)
  return ret1,ret2

def import_namespace(modulename,files, paths,imports,exports,deftypes):
  ret = ""
  head, tail = headtailcode(modulename, paths, imports, exports)
  ret = "__subshell = globals().copy()\n"
  ret += head % dict(shell="__subshell") + "\n"  
  for f in files:
    ret += "execfile(r\"%s\", __subshell)\n" % os.path.abspath(f)  
  for typ in deftypes:
    ret += "%s = __subshell[\"%s\"]\n" % (typ, typ)
    for n in range(Spyder.core.__arraycount[typ]):
      typ2 = typ + (n+1) * "Array"
      ret += "%s = __subshell[\"%s\"]\n" % (typ2, typ2)
  ret += tail % dict(shell="__subshell") + "\n"
  return ret
  
def systemcompile(lines, enabledswitches=set(),spyderdir=None,has_tempdir0=False):
  global has_tempdir, systempdir, namespaces, namespacelist, assesspar
  systempdir = None
  has_tempdir = has_tempdir0
  system = {}
  namespaces = {}
  switches = {"":[]}
  switchlist = ["",]
  toplevelfiles, switchvars = parse(lines, switches, switchlist,spyderdir)
  assesspar = general()
  assesspar.switches = switches
  assesspar.switchvars = switchvars
  assesspar.switchlist = switchlist
  assesspar.currswitch = "" 
  assesspar.enabled = enabledswitches
  assesspar.checked = set()                      
  assess_enabling(assesspar, spyderdir)
  l = []
  l += switches[""]
  for s in switchlist:
    if s in assesspar.enabled: l += switches[s]
  for s in enabledswitches:
    if s not in switchlist and s != "NOCOMPILE" and s != "SILENT":
      print >> sys.stderr, "WARNING: unknown switch %s" % s
  assemble(namespaces, l)
  namespacelist = []  
  for nr in range(len(namespaces)):
    minn = None
    minnr = 999999
    for n in namespaces:
      if n not in namespacelist and namespaces[n].order < minnr: 
        minnr = namespaces[n].order
        minn = n
    namespacelist.append(minn)
  #assert len(namespacelist) == len(namespaces)  
  for n in namespacelist:
    compile_core_namespace(system, namespaces[n])
  change = True  
  while change:
    change = False
    for n in namespacelist:
      if namespaces[n].compiled == False:
        compile_namespace(system, namespaces, namespaces[n])
        if namespaces[n].compiled == True:
          change = True
  for n in namespacelist:
    if namespaces[n].compiled == False:
      raise Exception("System error: cannot compile module %s due to the following undefined types:" % namespaces[n].name, list(namespaces[n].neededtypes))
  corecode = ""
  for n in namespacelist:
    if namespaces[n].iscore and len(namespaces[n].codefiles) > 0:
      corecode  += import_namespace(n, namespaces[n].codefiles, namespaces[n].paths, namespaces[n].imports, namespaces[n].exports, namespaces[n].defined) % dict(shell="currnamespace")
      continue
    determine_called_namespaces(system, namespaces, n)
        
  type_to_namespaces = {}
  for t in system:
    n = []
    if system[t].definition == None: directspaces = []
    else: directspaces = [system[t].definition,]
    if hasattr(system[t], "support"):
      directspaces += system[t].support
    for space in directspaces:
      for nam in namespaces[space].called:
        nr = namespacelist.index(nam)
        if (nam, nr) not in n:
          n.append((nam, nr))    
    type_to_namespaces[t] = n
  return corecode, type_to_namespaces, namespaces, toplevelfiles, systempdir
  
    
#TO DO: 
#- EXECORE and METACORE (METACORE to exec during compilation (Python),
# EXECORE to exec during runtime (target language)
# std CORE to do both (Python)
# this is a FILE attribute not a MODULE attribute
#- SWITCH to assume that the system did not change (no recompile)
